<?php

namespace Drupal\monarch_entity_yml\Commands;

use Drush\Commands\DrushCommands;
use Drupal\Core\Serialization\Yaml;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\menu_link_content\MenuLinkContentInterface;
use Drupal\path\Plugin\Field\FieldType\PathFieldItemList;
use Drupal\Core\Field\EntityReferenceFieldItemListInterface;

/**
 * A Drush commandfile.
 *
 * In addition to this file, you need a drush.services.yml
 * in root of your module, and a composer.json file that provides the name
 * of the services file to use.
 *
 * See these files for an example of injecting Drupal services:
 *   - http://cgit.drupalcode.org/devel/tree/src/Commands/DevelCommands.php
 *   - http://cgit.drupalcode.org/devel/tree/drush.services.yml
 */
class EntityYmlCommands extends DrushCommands {

  /**
   * Referenced allowed to be exported.
   *
   * @var array
   */
  protected $exportReferenceTypes = [
    'path_alias' => FALSE,
  ];

  /**
   * Explicitly export ids.
   *
   * @var array
   */
  protected $explicit = [];

  /**
   * Include non-banned references.
   *
   * @var bool
   */
  protected $includeReferences = FALSE;

  /**
   * Should this entity be exported?
   */
  protected function shouldExport(EntityInterface $entity) : bool {
    if (!($entity instanceof ContentEntityInterface)) {
      return FALSE;
    }

    if ($this->explicit[$entity->getEntityTypeId()][$entity->id()] ?? NULL) {
      return TRUE;
    }

    if ($this->includeReferences && !isset($this->exportReferenceTypes[$entity->getEntityTypeId()])) {
      return TRUE;
    }

    return $this->exportReferenceTypes[$entity->getEntityTypeId()] ?? FALSE;
  }

  /**
   * Write an entity to yml.
   */
  protected function writeEntity(ContentEntityInterface $entity, string $path, array &$done = [], array &$circular = []) {
    $uuid = $entity->uuid->value ?? NULL;

    if (empty($uuid) || ($circular[$uuid] ?? NULL)) {
      return NULL;
    }

    if ($done[$uuid]) {
      return $uuid;
    }

    $circular[$uuid] = $done[$uuid] = $uuid;

    $dependencies = [];

    foreach ($entity->getFields() as $field) {
      if ($field instanceof EntityReferenceFieldItemListInterface) {
        foreach ($field->referencedEntities() as $childEntity) {
          if ($childEntity && $childEntity instanceof ContentEntityInterface) {
            if ($this->shouldExport($childEntity)) {
              $dependencies[] = $this->writeEntity($childEntity, $path, $done);
            }
          }
        }
      }
    }

    if ($entity instanceof MenuLinkContentInterface) {
      $parent_id = $entity->getParentId();

      if (!empty($parent_id)) {
        $parent_id = explode(':', $parent_id);

        if (($parent_type = $parent_id[0] ?? NULL) && ($parent_uuid = $parent_id[1] ?? NULL)) {
          $parentEntity = \Drupal::service('entity.repository')->loadEntityByUuid($parent_type, $parent_uuid);

          if ($parentEntity && $parentEntity instanceof ContentEntityInterface) {
            if ($this->shouldExport($parentEntity)) {
              $dependencies[] = $this->writeEntity($parentEntity, $path, $done);
            }
          }
        }
      }
    }

    $values = [];

    foreach ($entity->getFields() as $key => &$field) {
      try {
        $values[$key] = $this->getFieldValue($field);
      }
      catch (\Throwable $err) {
        // Just skip adding the field.
      }
    }

    $filename_parts = [
      $entity->getEntityTypeId(),
      $entity->id(),
      $uuid,
    ];

    file_put_contents(realpath($path) . '/' . implode('__', $filename_parts) . '.yml', Yaml::encode([
      'storage' => $entity->getEntityTypeId(),
      'dependencies' => array_filter($dependencies),
      'values' => $values,
    ]));

    return $uuid;
  }

  /**
   * Export entity as configuration.
   *
   * @param string $entity_list
   *   Entity ids.
   * @param array $options
   *   The options provided by drush.
   *
   * @command entity_yml:dump
   *
   * @aliases eyd
   * @option $with-references Include all references (unless explicitly excluded).
   * @option $path Export yml files to this directory instead of searching for the correct directory.
   * @option $unexported List unexported references.
   * @usage entity_yml:dump entity_type:id
   *   Export an entity to yml.
   * @usage entity_yml:dump entity_type:id,id,id...
   *   Export multiple entities of a type to yml.
   * @usage entity_yml:dump entity_type:>=id...
   *   Export entities of a type equal to or greater than a specific id to yml.
   * @usage entity_yml:dump entity_type:<=id...
   *   Export entities of a type equal to or lesser than a specific id to yml.
   * @usage entity_yml:dump entity_type:*
   *   Export all entities of a type to yml.
   * @usage entity_yml:dump entity_type:+
   *   Export all secondary references to entities of a type to yml.
   * @usage entity_yml:dump entity_type:-
   *   Exclude entity type from export (but still check it for references, if
   *   included by other criteria).
   * @usage entity_yml:dump ...;...;...
   *   Specify multiple criteria.
   * @usage entity_yml:dump "... ... ..."
   *   Specify multiple criteria.
   * @usage entity_yml:dump "entity_type:id another_entity_type:* another_entity_type:+"
   *   Export multiple entities and their referenced entities to yml.
   */
  public function entityExport(string $entity_list = NULL, array $options = [
    'with-references' => FALSE,
    'path' => FALSE,
    'unexported' => FALSE,
  ]) {
    $entityTypeManager = \Drupal::service('entity_type.manager');

    $this->includeReferences = $options['with-references'] ? TRUE : FALSE;
    $this->exportReferenceTypes = [
      'path_alias' => FALSE,
    ];

    $queue = [];

    preg_match_all('/([^\s;:]+):([^\s;:]+)/i', $entity_list, $matches, PREG_SET_ORDER);

    foreach ($matches as $match) {
      try {
        $storage = $entityTypeManager->getStorage($match[1]);

        if (empty($storage)) {
          $this->stderr()->writeln('Entity storage not found: ' . $match[1]);
          continue;
        }

        if ($match[2] === '+') {
          $this->exportReferenceTypes[$match[1]] = $this->exportReferenceTypes[$match[1]] ?? TRUE;
        }
        elseif ($match[2] === '-') {
          $this->exportReferenceTypes[$match[1]] = FALSE;
        }
        elseif ($match[2] === '*') {
          $res = $storage->loadMultiple();
          foreach ($res as $entity) {
            $this->explicit[$match[1]][$entity->id()] = $entity;
          }
        }
        elseif ($match[2] === '+') {
          continue;
        }
        else {
          $ids = [];

          foreach (explode(',', $match[2]) as $idstr) {
            if (preg_match('/>=(.+)/', $idstr, $submatches)) {
              $idval = $submatches[1];

              if (is_numeric($idval)) {
                $idval = (int) $idval;

                foreach (($idlist = $storage->getQuery()->execute()) as $possible_id) {
                  if ($possible_id >= $idval) {
                    $ids[] = $possible_id;
                  }
                }
              }
            }
            elseif (preg_match('/<=(.+)/', $idstr, $submatches)) {
              $idval = $submatches[1];

              if (is_numeric($idval)) {
                $idval = (int) $idval;

                foreach (($idlist = $storage->getQuery()->execute()) as $possible_id) {
                  if ($possible_id <= $idval) {
                    $ids[] = $possible_id;
                  }
                }
              }
            }
            else {
              $ids[] = $idstr;
            }
          }

          $res = $storage->loadMultiple($ids);
          foreach ($res as $entity) {
            $this->explicit[$match[1]][$entity->id()] = $entity;
          }
        }

        foreach ($res as $entity) {
          if (!($entity instanceof ContentEntityInterface)) {
            $this->stderr()->writeln('Skipping non-content entity: ' . $entity->getEntityTypeId() . ':' . $entity->id());
            continue;
          }

          if (empty($entity)) {
            $this->stderr()->writeln('Entity not found: ' . $match[0]);
            continue;
          }

          $queue[] = $entity;
        }
      }
      catch (\Throwable $e) {
        $this->stderr()->writeln($e->getMessage());
      }
    }

    $fs = \Drupal::service('file_system');
    if (is_string($options['path'] ?? NULL)) {
      $content_path = $fs->realpath($options['path']);

      if (!$content_path) {
        $this->stderr()->writeln('Path not found. Please create the directory and ensure that it is writable.');
        return;
      }
    }
    else {
      $content_path = $fs->realpath(\Drupal::service('extension.path.resolver')->getPath('profile', \Drupal::installProfile()) . '/content')
        ?: $fs->realpath('../content')
        ?: $fs->realpath('content');

      if (!$content_path) {
        $this->stderr()->writeln('Content directory not found. Please create a \'content\' directory either in your install profile or at the root of your installation.');
        return;
      }
    }

    $done = [];

    foreach ($queue as $entity) {
      $this->writeEntity($entity, $content_path, $done);
    }
  }

  /**
   * Get field value.
   *
   * @param \Drupal\Core\Field\FieldItemListInterface $field
   *   The field to extract the value from.
   *
   * @return mixed
   *   The contents of the field.
   */
  protected function getFieldValue(FieldItemListInterface $field) {
    $ret = $field->getValue();

    if ($field instanceof PathFieldItemList && is_array($ret)) {
      foreach ($ret as &$values) {
        unset($values['pid']);
      }
    }

    return $ret;
  }

  /**
   * Export entity as configuration.
   *
   * @command entity_yml:import
   *
   * @aliases eyi
   * @option $path Load yml files from this directory instead of searching for the correct directory.
   * @usage entity_yml:import
   *  Imports yml and attempts to recreate entities defined within.
   */
  public function entityImport(array $options = [
    'path' => FALSE,
  ]) {
    $fs = \Drupal::service('file_system');

    if (is_string($options['path'] ?? NULL)) {
      $content_path = $fs->realpath($options['path']);

      if (!$content_path) {
        throw new \Exception('Path not found. Please create the directory and ensure that it is writable.');
      }
    }
    else {
      $content_path = $fs->realpath(\Drupal::service('extension.path.resolver')->getPath('profile', \Drupal::installProfile()) . '/content')
        ?: $fs->realpath('../content')
        ?: $fs->realpath('content');

      if (!$content_path) {
        throw new \Exception('Content directory not found. Please create a \'content\' directory either in your install profile or at the root of your installation.');
      }
    }

    try {
      $feedback = \Drupal::service('monarch_entity_yml.import')->import($content_path);

      $this->output()->writeln('Entities imported:');

      foreach ($feedback as $entry) {
        $this->output()->writeln($entry);
      }
    }
    catch (\Throwable $e) {
      $this->stderr()->writeln($e->getMessage());
      return;
    }
  }

  /**
   * List entity types eligible for export.
   *
   * @usage entity_yml:list
   *  List entities eligible for export.
   *
   * @command entity_yml:list
   * @aliases eyl
   */
  public function entityList() {
    $entityTypeManager = \Drupal::service('entity_type.manager');
    $entity_definitions = $entityTypeManager->getDefinitions();

    $list = [];

    foreach ($entity_definitions as $definition) {
      if ($definition instanceof ContentEntityTypeInterface) {
        $list[] = $definition->id();
      }
    }

    if (empty($list)) {
      $this->stderr()->writeln('No entity types eligible for export.');
    }
    else {
      $this->output()->writeln('Entity types eligible for export:');

      sort($list);

      foreach ($list as $entity_type) {
        $this->output()->writeln("\t" . $entity_type);
      }
    }
  }

}
